//*********************************************************************
//* mTouch CAPACITIVE SENSING
//* PIC16F727 VERSION
//* NUTS AND VOLTS MAGAZINE
//* AUTHOR:  FRED EADY
//* REV 1.0 05/18/08
//*********************************************************************

#include <pic.h>
#include <stdio.h>

__CONFIG(INTIO & WDTDIS & PWRTDIS & MCLREN & BORDIS & PLLEN & DEBUGEN);
__CONFIG(VCAPRA0);
__IDLOC(0518);

//******************************************************************
//	KEYPAD BIT STRUCTURE
//******************************************************************
typedef struct 
	{
		char PAD0 : 1;
		char PAD1 : 1;
		char PAD2 : 1;
		char PAD3 : 1;
		char PAD4 : 1;
		char PAD5 : 1;
		char PAD6 : 1;
		char PAD7 : 1;
		char PAD8 : 1;
		char PAD9 : 1;
		char PAD10 : 1;
		char PAD11 : 1;
	} PPad;
	
	PPad keypad;
//******************************************************************
//	SENSOR COUNT DEFINITIONS
//******************************************************************
#define	MAXSENSORS		12
#define	NUM_AVG_PASSES	3
#define NUM_STAB_PASSES 5
//******************************************************************
//    MACROS AND PORT/PIN DEFINITIONS
//******************************************************************
#define make8(var,offset)	((unsigned int)var >> (offset * 8)) & 0x00FF
#define	make16(varhigh,varlow)	(((unsigned int)varhigh & 0xFF)* 0x100) + ((unsigned int)varlow & 0x00FF)
//*********************************************************************
//*	FUNCTION PROTOTYPES
//*********************************************************************
void init(void);
void interrupt SENSORS(void);
void reset_timer1(void);
void select_next_button(void);
//******************************************************************
//*	I/O DEFINITIONS
//******************************************************************
#define	btn_LED			RC0
#define	ON				0		// LED ON 
#define OFF				1		// LED OFF
//******************************************************************
//*	SRAM DEFINITIONS
//******************************************************************
bank2 unsigned int	timer1_val[MAXSENSORS];
bank2 unsigned int	timer1_raw[MAXSENSORS];
unsigned int	btn_average[MAXSENSORS];
unsigned int	trip_val[MAXSENSORS];
char			stabilize_pass;
char			average_pass;
char			btn_pressed;
char			sensor_index;
//*********************************************************************
//*	INTERRUPT DEFINITIONS
//*********************************************************************
#define		enable_GLOBALint  	GIE = 1;				\
							  	PEIE = 1;
#define		disable_GLOBALint 	GIE = 0;				\
							  	PEIE = 0;
//*********************************************************************
//*	TIMER MACROS
//*********************************************************************
#define		TIMER0ON			TMR0ON = 1
#define		TIMER0OFF			TMR0ON = 0
#define		TIMER1ON			TMR1ON = 1
#define		TIMER1OFF			TMR1ON = 0
#define		TIMER2ON		  	TMR2ON = 1
#define		TIMER2OFF		  	TMR2ON = 0
#define		TIMER3ON		  	TMR3ON = 1
#define		TIMER3OFF		  	TMR3ON = 0
//*********************************************************************
//*	putch FUNCTION FOR PRINTF
//*********************************************************************
void putch(char c)
{
   while(!TRMT) 	// TRMT is set when TSR is empty
      continue;

  TXREG = c;		// load the register
}
//*********************************************************************
//*	TIMER1 RESET FUNCTION
//*********************************************************************
void reset_timer1(void) 
{
	TMR1L	= 0;	// clear TIMER1 low byte
	TMR1H	= 0;	// clear TIMER1 high byte
	TIMER1ON;		// start TIMER1
}
//*********************************************************************
//*	SELECT NEXT CPS CHANNEL
//*********************************************************************
void select_next_sensor(void)
{
	++sensor_index;
	if(sensor_index >= MAXSENSORS)
	{
		sensor_index = 0;
	}
	CPSCON1 = sensor_index;
}
//*********************************************************************
//*	TOUCH INTERRUPT HANDLER
//*********************************************************************
void interrupt SENSORS(void)
{
	
	if(TMR1GIF && TMR1GIE)
	{
		TMR1GIF = 0;
		TIMER1OFF;
		
		timer1_val[sensor_index] = make16(TMR1H,TMR1L);
		timer1_raw[sensor_index] = timer1_val[sensor_index] * 16;
		
		//stabilize the btn_average value 
		if(stabilize_pass > 0x00)
		{
			--stabilize_pass;
			btn_average[sensor_index] = timer1_raw[sensor_index];
			reset_timer1();
			
			if((stabilize_pass == 0) && (sensor_index < (MAXSENSORS - 1)))
			{
				stabilize_pass = 10;
				select_next_sensor();
			}
			
		}
		else
		{
			//check for button pressed
			if(timer1_raw[sensor_index] < (btn_average[sensor_index] - trip_val[sensor_index]))
			{
				if(!(keypad.PAD0 || keypad.PAD1 || keypad.PAD2 ||
					 keypad.PAD3 || keypad.PAD4 || keypad.PAD5 ||
					 keypad.PAD6 || keypad.PAD7 || keypad.PAD8 ||
					 keypad.PAD9 || keypad.PAD10 || keypad.PAD11))
					 {
						 switch(sensor_index)
						 {
							 case 0:	keypad.PAD0 = 1;	break;
							 case 1:	keypad.PAD1 = 1;	break;
							 case 2:	keypad.PAD2 = 1;	break;
							 case 3:	keypad.PAD3 = 1;	break;
							 case 4:	keypad.PAD4 = 1;	break;
							 case 5:	keypad.PAD5 = 1;	break;
							 case 6:	keypad.PAD6 = 1;	break;
							 case 7:	keypad.PAD7 = 1;	break;
							 case 8:	keypad.PAD8 = 1;	break;
							 case 9:	keypad.PAD9 = 1;	break;
							 case 10:	keypad.PAD10 = 1;	break;
							 case 11:	keypad.PAD11 = 1;	break;
						}
					}
					 
			}
			else if(timer1_raw[sensor_index] > (btn_average[sensor_index] - trip_val[sensor_index])+30)
			{
						 switch(sensor_index)
						 {
							 case 0:	keypad.PAD0 = 0;	break;
							 case 1:	keypad.PAD1 = 0;	break;
							 case 2:	keypad.PAD2 = 0;	break;
							 case 3:	keypad.PAD3 = 0;	break;
							 case 4:	keypad.PAD4 = 0;	break;
							 case 5:	keypad.PAD5 = 0;	break;
							 case 6:	keypad.PAD6 = 0;	break;
							 case 7:	keypad.PAD7 = 0;	break;
							 case 8:	keypad.PAD8 = 0;	break;
							 case 9:	keypad.PAD9 = 0;	break;
							 case 10:	keypad.PAD10 = 0;	break;
							 case 11:	keypad.PAD11 = 0;	break;
						}
			}
			
			if(timer1_raw[sensor_index] > (btn_average[sensor_index]))
			{
				btn_average[sensor_index] = btn_average[sensor_index] + (timer1_val[sensor_index] - (btn_average[sensor_index] / 16));
			}
			//compute average
			if(--average_pass == 0)
			{
				btn_average[sensor_index] = btn_average[sensor_index] + (timer1_val[sensor_index] - (btn_average[sensor_index] / 16));
				average_pass = NUM_AVG_PASSES;
				select_next_sensor();
			}
			reset_timer1();
		}
	}
}

//*********************************************************************
//*	INITIALIZE FUNCTION
//*********************************************************************
void init(void)
{
	char x;
	
	OSCCON = 0x30;	//16 MHz
//*********************************************************************
//*	SETUP I/O PORTS
//*********************************************************************
	TRISA = 0b10111111;		//PARTIALLY AVAILABLE
	TRISB = 0b11111111;		//CPS USED
	TRISC = 0b11111110;		//PARTIALLY AVAILABLE
	TRISD = 0b11111111;		//CPS PARTIALLY USED
	TRISE = 0b11111111;		//AVAILABLE
	
	ANSELB = 0b00111111;	//PORTB = ANALOG
	ANSELD = 0b11111111;	//PORTD = ANALOG
	ANSELE = 0;				//PORTE = DIGITAL
//*********************************************************************
//*	SETUP TIMERS
//*********************************************************************
	OPTION = 0b11000101;	//TIMER0 CLK ON FOSC/4
	T1CON =  0b01000101;
	T1GCON = 0b11100001;
//*********************************************************************
//*	SETUP CAP SENSE MODULE
//*********************************************************************
	CPSCON0 = 0b10001100;
	CPSCON1 = 0;
//*********************************************************************
//*	SETUP CAP SENSE VARIABLES
//*********************************************************************
	stabilize_pass = NUM_STAB_PASSES;
	average_pass = NUM_AVG_PASSES;
	
	trip_val[0] = 120;
	trip_val[1] = 120;
	trip_val[2] = 120;
	trip_val[3] = 120;
	trip_val[4] = 120;
	trip_val[5] = 120;
	trip_val[6] = 120;
	trip_val[7] = 120;
	trip_val[8] = 120;
	trip_val[9] = 120;
	trip_val[10] = 120;
	trip_val[11] = 120;
	
	for(x = 0; x < MAXSENSORS; ++x)
	{
		btn_average[x] = 0;
	}
		sensor_index = 0;
//*********************************************************************
//*	SETUP AUSART
//*********************************************************************
	SPBRG = 16;	//56Kbps
	BRGH = 1;
	SPEN = 1;	//enable AUSART
	CREN = 1;	//enable RX		
	TXEN = 1;	//enable TX
//*********************************************************************
//*	SETUP INTERRUPTS
//*********************************************************************
	TMR1GIF = 0;
	TMR1GIE = 1;
	enable_GLOBALint;
}	

void main(void) 
{
	init();	
	
		// signal button status
	do{
		if(keypad.PAD0)
		{
			btn_LED = ON;
			printf("PAD0\r\n");
		}
		else
		{
			btn_LED = OFF;
		}
		
		if(keypad.PAD1)
		{
			btn_LED = ON;
			printf("PAD1\r\n");
		}
		else
		{
			btn_LED = OFF;
		}

		if(keypad.PAD2)
		{
			btn_LED = ON;
			printf("PAD2\r\n");
		}
		else
		{
			btn_LED = OFF;
		}

		if(keypad.PAD3)
		{
			btn_LED = ON;
			printf("PAD3\r\n");
		}
		else
		{
			btn_LED = OFF;
		}
		
		if(keypad.PAD4)
		{
			btn_LED = ON;
			printf("PAD4\r\n");
		}
		else
		{
			btn_LED = OFF;
		}

		if(keypad.PAD5)
		{
			btn_LED = ON;
			printf("PAD5\r\n");
		}
		else
		{
			btn_LED = OFF;
		}
							
		if(keypad.PAD6)
		{
			btn_LED = ON;
			printf("PAD6\r\n");
		}
		else
		{
			btn_LED = OFF;
		}
		
		if(keypad.PAD7)
		{
			btn_LED = ON;
			printf("PAD7\r\n");
		}
		else
		{
			btn_LED = OFF;
		}

		if(keypad.PAD8)
		{
			btn_LED = ON;
			printf("PAD8\r\n");
		}
		else
		{
			btn_LED = OFF;
		}

		if(keypad.PAD9)
		{
			btn_LED = ON;
			printf("PAD9\r\n");
		}
		else
		{
			btn_LED = OFF;
		}
		
		if(keypad.PAD10)
		{
			btn_LED = ON;
			printf("PAD10\r\n");
		}
		else
		{
			btn_LED = OFF;
		}

		if(keypad.PAD11)
		{
			btn_LED = ON;
			printf("PAD11\r\n");
		}
		else
		{
			btn_LED = OFF;
		}

	}while(1);

}

